'''


    Messages
    --------

    There are two primitives:


        propose(dcaddress, string, [t0, t1])
        dismiss(dcaddress, string, [t0, t1])


    where:

        dcaddress = [bucket1, bucket2, ...]
        validity = [from, to] or [null, to] or [from, null]

    The semantics of

        propose([bucket1, bucke2], string, [t0, t1])
        state['bucket1']['bucket2'].add(string)

        so that it is true that

            string in state['bucket1']['bucket2'] for t0 < t < t1

    This is true, unless a "dismiss" or "replace" command is given.

    Summaries
    ---------

    There is a message:

        ask-summary
        summary(ipfs)

    where ipfs is the multihash of a list of YAML messages to be interpreted.


    Signatures
    -------

        {mtype:'signed', what='QmContent', peerid='QmPublicKeyHash', signature='...'}

    Envelope
    -------

    There is some meta-information associated:

        signatures: a list of signatures



    Signature
    ---------

    A signature has:

        ID: (hash of public key)
        Signature: bytes
        Validity: [t0, t1]

    Routing
    -------

    channels: channels in which this was circulated, or None



    Python API querying
    ========

    In Python, a dcaddress is represented by a tuple of strings.

        dcaddress = ('my', 'prefix')

    A client is able to query as follows:

        h = storage.dca(dcaddress)

    List the contents

        ss = storage.list(dcprefix)

        for s in sl:
            s.data        # data
            s.validity    # t0, inf
            s.channels    # channels this was received from
            s.signatures  # collected signatures

    Distributed authenticated sets

        peer [PeerID]  The host HASH is a Duckiebot

            propose: peer
            dismiss: admin

        trusted [PeerID]

            propose: admin
            dismiss: admin

        admin [PeerID]

            propose: mama
            dismiss: mama

        summary [Hash]

            propose: anybody
            dismiss: trusted

        networks

            propose: admin
            dismiss: admin

            examples:

                /irc/address
                /udp/address

        files / HASH

            propose: anybody
            dismiss: trusted

        uploads / HASH

            propose: trusted
            dismiss: trusted

        propose('files', ipfs)
        dismiss('files', ipfs)

        propose('valid', ipfs)
        dismiss('valid', ipfs)
        replace('valid', ifps1, ipfs2)

        propose('summary', H('irc1'))

        propose('network', H('irc1'))
        propose('network', H('irc1'))

        propose(ipfs_host, 'role')



    Addresses

    /duckieswarm/<ID>/verifier
    /duckieswarm/<ID>/brain

    /dns4/frankfurt.co-design.science/tcp/6667/irc/#duckiebots
    /dns4/frankfurt.co-design.science/tcp/6667/irc/NickName
    /ip4/10.0.0.0/udp/6060/duckieswarmcall

'''

from collections import defaultdict, OrderedDict
from duckietown_swarm.intervals_utils import interval_friendly, in_interval
from duckietown_swarm.proposals import Proposal, strictly_dominated_by_set, \
    ANONYMOUS
from duckietown_swarm.utils import get_sha256_base58
import itertools
import json
import os
import shelve
import time

from base58 import b58encode, b58decode
from contracts import contract
from contracts.utils import raise_wrapped, indent, raise_desc

from .dcache_wire import CouldNotInterpret, interpret_message, MsgPropose
from .dcache_wire import MsgDismiss, MsgSignature, create_propose_message
from .dcache_wire import MsgPing, MsgPong
from .dcache_wire import create_dismiss_message
from .dcache_wire import create_signature_message
from .utils import pretty_print_dictionary, friendly_time_since

BOOTSTRAP_PATTERNS = ['/peers', '*/summaries', '*/ipfs_addresses', '*/info/*', '*/versions/*']


class Envelope(object):

    @staticmethod
    def encode(s):
#        return base92.b92encode(s)
#        return base85.b85encode(s)
        return s
        return b58encode(s)

    @staticmethod
    def decode(s):
        return s
        return b58decode(s)
#        return base85.b85decode(s)
#        return base92.b92decode(s)

    def __init__(self, maddr_from, maddr_to, contents, comment=None):
        self.maddr_from = maddr_from
        self.maddr_to = maddr_to
        self.contents = contents
        self.comment = comment

    def __repr__(self):
        return 'Envelope(from: %s, to: %s, %s, %s)' % (self.maddr_from if self.maddr_from else '@',
                                                   self.maddr_to if self.maddr_to else '@',
                                                   self.comment if self.comment else '(nc)',
                                                   self.contents)

    def verbose(self):
        od = OrderedDict()
        od['from'] = self.maddr_from.__repr__()
        od['to'] = self.maddr_to.__repr__()
        od['payload'] = self.contents.__repr__()
        if self.comment:
            od['comment'] = self.comment.__repr__()
        return 'Envelope' + '\n' + indent(pretty_print_dictionary(od), '  ')

    def to_json(self):
        od = OrderedDict()
        od['from'] = self.maddr_from
        od['to'] = self.maddr_to
        od['payload'] = Envelope.encode(self.contents)
        if self.comment:
            od['comment'] = self.comment
        s = json.dumps(od)
        if len(s) > 512:
            print('warning: size of message is %s' % len(s))
        return s

    @staticmethod
    def from_json(s):
        if not s.startswith('{') or not s.endswith('}'):
            msg = 'The message of length %s appears truncated.' % len(s)
            raise_desc(CouldNotReadEnvelope, msg, s=s.__repr__())
        try:
            od = json.loads(s)
            payload = Envelope.decode(str(od['payload']))
            comment = od.get('comment', None)
            return Envelope(str(od['from']), str(od['to']), payload, comment=comment)
        except Exception as e:
            msg = 'Could not read envelope of length %s.' % len(s)
            raise_wrapped(CouldNotReadEnvelope, e, msg, s=s)


class CouldNotReadEnvelope(Exception):
    pass


class NoSuchBucket(Exception):
    pass


def expired(validity, t=None):
    if t is None:
        t = time.time()
    _, t1 = validity
    if t1 is None:
        return False
    return t > t1


class BucketDataItem(object):

    def __init__(self):
#        self.validity2signatures_propose = defaultdict(set)
        self.validity2signatures_dismiss = defaultdict(set)

        self.proposals_propose = set()

        self.last_broadcast_propose = {}
        self.last_broadcast_dismiss = {}

    @contract(validity='seq[2]')
    def propose(self, validity, signatures, from_channel):
#        print('BucketDataItem %s %s %s' % (validity, signatures, from_channel))
        if not signatures:
            signatures = [ANONYMOUS]
        for s in signatures:
            p = Proposal(validity, s)
            self.propose_(p, from_channel)

    @contract(proposal=Proposal)
    def propose_(self, proposal, from_channel):
        if proposal in self.proposals_propose:
            return
#        else:
#            print('%s is not in %s ' % (proposal, self.proposals_propose))
        dominated_by = strictly_dominated_by_set(proposal, self.proposals_propose)
        if dominated_by:
            return
#            msg = 'Ignoring proposed:'
#            msg += '\n\n' + indent(proposal, '  ')
#            msg += '\n\nBecause dominated by'
#            for d in dominated_by:
#                msg += '\n\n' + indent(d, '  ')
#            print msg
#            return dominated_by

        self.proposals_propose.add(proposal)

#        self.validity2signatures_propose[proposal.validity].update(proposal.signatures)

        if not from_channel in self.last_broadcast_propose:
            self.last_broadcast_propose[from_channel] = Stats()

        self.last_broadcast_propose[from_channel].just_received()

        _cleaned = self.cleanup()

#        if cleaned:
#            msg = 'Cleaning results:\n\n' + indent("\n".join(cleaned), '  ')
#            print msg

#        return cleaned

    def dismiss(self, validity, signatures, from_channel):
        self.validity2signatures_dismiss[validity].update(signatures)
        if not from_channel in self.last_broadcast_dismiss:
            self.last_broadcast_dismiss[from_channel] = Stats()
        self.last_broadcast_dismiss[from_channel].just_received()

#    def is_strictly_dominated(self, proposal):
#        available = set(self.iterate_existing_proposals())
#        return strictly_dominated_by_set(proposal, available)

#    def iterate_existing_proposals(self):
#        for v0, signers in self.validity2signatures_propose.items():
#            yield Proposal(v0, signers)
##            if signers:
#                for s in signers:
#                    yield Proposal(v0, [s])
#            else:
#                yield Proposal(v0, [])

    def cleanup_expired(self, t=None):
        for proposal in list(self.proposals_propose):
            if expired(proposal.validity, t):
                self.proposals_propose.remove(proposal)

        for validity in list(self.validity2signatures_dismiss):
            if expired(validity, t):
                self.validity2signatures_dismiss.pop(validity)

    def cleanup(self):
        cleaned_up = []
        # find all proposals
        dominated = defaultdict(set)
        current = list(self.proposals_propose)
        for c1, c2 in itertools.product(current, current):
            if Proposal.strictly_dominates(c1, c2):
                dominated[c2] = c1

        for proposal, by_what in dominated.items():
            cleaned_up.append('found redundant %s dominated by %s' % (proposal, by_what))
            self.proposals_propose.remove(proposal)
        return cleaned_up

    def is_tombstoned(self):
#        if len(self.validity2signatures_propose) == 0:
        for validity, _dataitem in self.validity2signatures_dismiss.items():
            if in_interval(validity, time.time()) and validity[1] == None:
                return True
        return False

    def summary(self):
        res = []

        if self.is_tombstoned():
            res.append('TOMBSTONED')
#        else:
#            res.append('ACTIVE')

        for proposal in self.proposals_propose:
            msg = 'propose %s by %s' % (interval_friendly(proposal.validity), proposal.who)
            res.append(msg)
        for interval, signatures in self.validity2signatures_dismiss.items():
            msg = 'dismiss %s by %s' % (interval_friendly(interval), signatures)
            res.append(msg)

        for channel, last in self.last_broadcast_propose.items():
            res.append(' %s: %s' % (last, channel))
        for channel, last in self.last_broadcast_dismiss.items():
            res.append('%s: %s' % (last, channel))

        return "\n".join(res)

    def valid(self, at):
        for validity, _signatures in self.validity2signatures_dismiss.items():
            if in_interval(validity, at):
                return False
        for proposal in self.proposals_propose:
            if in_interval(proposal.validity, at):
                return True
        return False


class Bucket(object):

    def __init__(self):
        self.children = OrderedDict()
        self.data2dataitem = defaultdict(BucketDataItem)
        self.hook_proposed = []
        self.hook_dismissed = []

    def __getstate__(self):
        return {'children': self.children,
                'data2dataitem': self.data2dataitem,
                'hook_proposed': [],
                'hook_dismissed': []}

    def cleanup(self):
        cleaned_up = []
        for data, dataitem in list(self.data2dataitem.items()):
            for _ in dataitem.cleanup():
                cleaned_up.append('%s | %s' % (data, _))

            if (not dataitem.proposals_propose and
                not dataitem.validity2signatures_dismiss):
                del self.data2dataitem[data]

        return cleaned_up

    def summary(self):

        res = OrderedDict()

        if self.children:
            od = OrderedDict()
            for k, v in self.children.items():
                od[k] = v.summary()

            res['children:'] = pretty_print_dictionary(od)

        if self.data2dataitem:
            od = OrderedDict()
            ntomb = 0

            for data, dataitem in self.data2dataitem.items():
                if dataitem.is_tombstoned():
                    ntomb += 1
                else:
                    od[data.__repr__()] = dataitem.summary()

            res['data:'] = pretty_print_dictionary(od)
            res['num tombstoned:'] = str(ntomb)

        return pretty_print_dictionary(res)

    @contract(validity='seq[2]')
    def propose(self, data, validity, roles, from_channel):
        found = data in self.data2dataitem
        ditem = self.data2dataitem[data]
        ditem.propose(validity, roles, from_channel)

        if not found:
            for hook in self.hook_proposed:
                hook(dict(data=data, validity=validity, roles=roles, from_channel=from_channel))

    def dismiss(self, data, validity, roles, from_channel):
        ditem = self.data2dataitem[data]
        ditem.dismiss(validity, roles, from_channel)

        for hook in self.hook_dismissed:
            hook(dict(data=data, validity=validity, roles=roles, from_channel=from_channel))

    def query(self, at):
        found = []
        for data, dataitem in self.data2dataitem.items():
            if dataitem.valid(at):
                found.append(data)
        return found

    def add_child_bucket(self, name):
        self.children[name] = Bucket()

    def items(self):
        for k, child in self.children.items():
            yield (k,), child
            for n1, c1 in child.items():
                yield (k,) + n1, c1

    @contract(add='str|(tuple,seq(str))')
    def get_bucket(self, add, create_if_not_exists):
        if isinstance(add, str):
            add = (add,)
        try:
            return self._get_bucket(add, create_if_not_exists)
        except KeyError as e:
            msg = 'Could not resolve bucket "%s": invalid key %s.' % ("/".join(add), e)
            msg += '\n create_if_not_exists: %s' % create_if_not_exists
            raise_wrapped(NoSuchBucket, e, msg)

    def _get_bucket(self, add, create_if_not_exists):
        if not add:
            return self
        name = add[0]
        rest = add[1:]
        if create_if_not_exists and (not name in self.children):
            self.children[name] = Bucket()
        return self.children[name]._get_bucket(rest, create_if_not_exists)


class ProcessFailure(Exception):
    pass


class Stats():

    def __init__(self):
        self.num_sent = 0
        self.num_received = 0
        self.last_sent = 0
        self.last_received = 0

    def get_last_activity(self):
        return max(self.last_sent, self.last_received)

    def just_received(self):
        self.num_received += 1
        self.last_received = time.time()

    def just_sent(self):
        self.num_sent += 1
        self.last_sent = time.time()

    def __repr__(self):
        s = []
        if self.num_sent == 0:
            s.append('')
        else:
            s.append('sent %3d  %s' % (self.num_sent, friendly_time_since(self.last_sent)))

        if self.num_received == 0:
            s.append('')
        else:
            s.append('recv %3d  %s' % (self.num_received, friendly_time_since(self.last_received)))

        s = [_.ljust(23) for _ in s]
        return " | ".join(s)


storage_version = 6


class DistributedRepo(object):

    get_data_hash = get_sha256_base58

    def __init__(self):
        self.root = Bucket()

        self.signatures = defaultdict(dict)
        self.cache_dir = None
        self.all_channels = defaultdict(Stats)

    def use_cache_dir(self, cache_dir):
        fname = os.path.join(cache_dir, '.cache.shelve')
        print('using cache: %s' % fname)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        self.shelf = shelve.open(fname, writeback=True)

        try:
            self._init()
        except Exception as e:
            print(e)
            print('Resetting cache')
            os.unlink(fname)
            self.shelf = shelve.open(fname, writeback=True)
            self._init()

    def close(self):
        self.sync()
        self.shelf.close()

    def sync(self):
        self.shelf['root'] = self.root
        self.shelf.sync()

    def _init(self):
        if not 'version' in self.shelf or self.shelf['version'] != storage_version:
            print('Resetting cache because of version')
            for f in ['root', 'signatures']:
                if f in self.shelf:
                    self.shelf.pop(f)

        self.shelf['version'] = storage_version

        self.shelf['root'] = self.root = self.shelf.get('root', self.root)
        self.shelf['signatures'] = self.signatures = self.shelf.get('signatures', self.signatures)

    def _get_roles_endorsing(self, wm, at):
        signatures = self._get_signatures(wm)
        roles = set()
        for s in signatures:
            roles.update(self._roles_from_signatures(s, at))
        if not roles:
            roles = set(['unknown'])
        return roles

    def _get_signatures(self, msg):
        hashed = get_sha256_base58(msg)
        if hashed in self.signatures:
            signatures = self.signatures[hashed]
            return list(signatures)
        else:
            return []

    def _roles_from_signatures(self, peerid, at=None):
        if at is None:
            at = time.time()
        roles = ['admin', 'trusted', 'peer']
        result = set()
        for role in roles:
            bucket = self.root.get_bucket(role, create_if_not_exists=False)
            contents = bucket.query(at)
            if peerid in contents:
                result.add(role)
        if not result:
            result = ['unknown']
        return set(result)

    def _add_signature(self, signer, data, signature):
        # todo: validate
#        expect = get_sha256_base58("")
#        if len(data) != len(expect):
#            msg = 'Invalid hash %s' % expect.__repr__()
#            raise ValueError(msg)
        self.signatures[data][signer] = signature

    def add_hook_proposed_all(self, f):

        class Wrapper():

            def __init__(self, name, f0):
                self.name = name
                self.f0 = f0

            def __call__(self, d):
                d['bucket_name'] = self.name
                self.f0(d)

        for bucket_name, bucket in self.root.items():
            bucket.hook_proposed.append(Wrapper(bucket_name, f))

    def add_hook_proposed(self, buckets, f):
        if isinstance(buckets, str):
            buckets = (buckets,)
        bucket = self.root.get_bucket(buckets, create_if_not_exists=False)
        bucket.hook_proposed.append(f)

    def add_hook_dismissed(self, buckets, f):
        if isinstance(buckets, str):
            buckets = (buckets,)
        bucket = self.root.get_bucket(buckets, create_if_not_exists=False)
        bucket.hook_dismissed.append(f)

    def summary(self):
        od = OrderedDict()
#        if False:
#            if self.signatures:
#                s = []
#                for data_hash, signatures in self.signatures.items():
#                    s.append('%s: %s' % (data_hash, signatures))
#                od['signatures'] = "\n".join(s)
        od['buckets'] = self.root.summary()

        s = []
        ordered = sorted(self.all_channels, key=lambda k:-self.all_channels[k].last_received)
        for channel in ordered:
            stats = self.all_channels[channel]
            s.append('%10s: %s' % (stats, channel))
        od['activity'] = "\n".join(s)

        return pretty_print_dictionary(od)

    def summary_messages(self):
        ''' Returns a list of strings that represent
            the data so far.
        '''

        out = []

        for bucket_name, _bucket in self.root.items():
            out.extend(self._summary_message_for_bucket(bucket_name))
        return out

    def summary_dict(self):
        ''' Returns a view of the data using simple structures '''

        @contract(D=dict)
        def setit(D, name, what):
            one, rest = name[0], name[1:]
            if not one in D: D[one] = {}
            if not rest:
                D[one] = what
            else:
                a = D[one]
                if not isinstance(a, dict):
                    raise ValueError('a = %s' % str(a))
                setit(a, rest, what)

        d = {}
        for bucket_name, bucket in self.root.items():
            datas = bucket.query(at=time.time())
            what = []
            for data in datas:
                dataitem = bucket.data2dataitem[data]
                if dataitem.is_tombstoned():
                    continue
                what.append(data)

            if what:
                try:
                    setit(d, bucket_name, what)
                except ValueError as e:
                    msg = 'Cannot set value for %s = %s: %s' % (bucket_name, what, e)
                    print(msg)
#                    raise_wrapped(ValueError, e, msg)

        return d

    def _summary_message_for_bucket(self, bucket_name, only_positive=False):
        out = []

        def add(msg):
            data_hash = get_sha256_base58(msg)
            available = self.signatures.get(data_hash, {})
#            if not available:
#                print('no signatures found for %s' % msg)
            for signer, signature in available.items():
                smsg = create_signature_message(data_hash, signer, signature)
                out.append(smsg)
#                print('Found signature by %s ' % (signer))
            out.append(msg)

        bucket = self.root.get_bucket(bucket_name, create_if_not_exists=False)
        datas = bucket.query(at=time.time())
        for data in datas:
            dataitem = bucket.data2dataitem[data]
            if dataitem.is_tombstoned():
                continue
            t = time.time()
            if not only_positive:
                for validity, _ in dataitem.validity2signatures_dismiss.items():
                    if validity[1] is None or t <= validity[1]:
                        msg = create_dismiss_message(bucket_name, data, validity)
                        add(msg)

            for proposal in dataitem.proposals_propose:
                if not expired(proposal.validity):
                    msg = create_propose_message(bucket_name, data, proposal.validity)
                    add(msg)
        return out

    @contract(buckets='seq(str)')
    def init_bucket(self, buckets):
        if isinstance(buckets, str):
            buckets = (buckets,)
        buckets = tuple(buckets)
        prev = buckets[:-1]
        name = buckets[-1]
        parent = self.root.get_bucket(prev, create_if_not_exists=True)
        parent.add_child_bucket(name)
#        print('initialized bucket %s' % "/".join(buckets))

    @contract(wm=str)
    def process(self, wm, from_channel, at=None):
        # print('process from %s: %s' % (from_channel, wm))
        if at is None:
            at = time.time()
        self.all_channels[from_channel].just_received()
        try:
            m = interpret_message(wm)
        except CouldNotInterpret as e:
            msg = 'Could not process message.'
            msg += '\n\n' + indent(wm, '  ') + '\n'
            raise_wrapped(ProcessFailure, e, msg, compact=True)

        try:
            if isinstance(m, MsgPropose):
                buckets = m.buckets

                data = m.data
                validity = m.validity
                bucket = self.root.get_bucket(buckets, create_if_not_exists=True)
#                roles = self._get_roles_endorsing(wm, at)
                signatures = self._get_signatures(wm)

                if buckets[0] == 'pri' and not (buckets[1] in signatures):
                    msg = 'Not setting %s because signatures=%s (from: %s)' % (buckets, signatures, from_channel)
                    print(msg)
                else:
                    bucket.propose(data, validity, signatures, from_channel)
            elif isinstance(m, MsgDismiss):
                buckets = m.buckets
                data = m.data
                validity = m.validity
                bucket = self.root.get_bucket(buckets, create_if_not_exists=True)
                signatures = self._get_signatures(wm)

                if buckets[0] == 'pri' and not buckets[1] in signatures:
                    msg = 'Not dismissing %s because signatures=%s (from: %s)' % (buckets, signatures, from_channel)
                    print(msg)
                else:
                    bucket.dismiss(data, validity, signatures, from_channel)
            elif isinstance(m, MsgSignature):
                self._add_signature(signer=m.signer, data=m.data, signature=m.signature)
            elif isinstance(m, MsgPing):
                print('ping %s' % from_channel)
            elif isinstance(m, MsgPong):
                print('pong %s' % from_channel)
            else:
                msg = 'Could not process class "%s".' % type(m).__name__
                msg += '\n\n' + indent(wm, '  ') + '\n'
                raise ProcessFailure(msg)
        except NoSuchBucket as e:
            msg = 'Could not find bucket.'
            msg += '\n\n' + indent(wm, '  ') + '\n'
            raise_wrapped(ProcessFailure, e, msg, compact=True)

    def query(self, bucket, at=None):
        '''
            Iterates when
        '''
        if at is None:
            at = time.time()
        if isinstance(bucket, str):
            bucket = (bucket,)
        bucket = self.root.get_bucket(bucket, create_if_not_exists=False)
        return bucket.query(at)

    def cleanup(self):
        for _bucket_name, bucket in self.root.items():
            removed = bucket.cleanup()
            if removed:
                print('For %s, removed\n%s' % (_bucket_name, removed))

    @contract(returns='seq($Envelope)')
    def rebroadcast(self, bucket_name, channels, min_delta=30, min_silence=5, max_n=1):
        from duckietown_swarm.brain import ALL_BRAINS
        options = []

        bucket = self.root.get_bucket(bucket_name, create_if_not_exists=False)
        datas = bucket.query(at=time.time())
        for data in datas:
            dataitem = bucket.data2dataitem[data]
            for channel in channels:
                last_activity = self.all_channels[channel].get_last_activity()

                last = dataitem.last_broadcast_propose.get(channel, Stats()).get_last_activity()
                enough_silence = time.time() - last_activity > min_silence
                if not enough_silence:
                    continue
                need_update = time.time() - last > min_delta
                if need_update:
                    for proposal in dataitem.proposals_propose:
                        validity = proposal.validity
#                        validity = list(dataitem.validity2signatures_propose)[0]
                        # XXX: too many?
                        msg = create_propose_message(bucket_name, data, validity=validity)
                        options.append((last, Envelope('', channel + ALL_BRAINS, msg), data, channel))

        s = [_[1:] for _ in sorted(options, key=lambda _: _[0])]
        s = s[:max_n]

        for _envelope, data, channel in s:
            dataitem = bucket.data2dataitem[data]
            if not channel in dataitem.last_broadcast_propose:
                dataitem.last_broadcast_propose[channel] = Stats()
            dataitem.last_broadcast_propose[channel].just_sent()
            self.all_channels[channel].just_sent()

        return [_[0] for _ in s]


class ChannelStats(object):

    @contract(last_seen=float, signatures=set)
    def __init__(self, last_seen, signatures):
        self.last_seen = last_seen
        self.signatures = signatures




from duckietown_swarm.intervals_utils import interval_friendly
import time
from contracts import contract

ANONYMOUS = 'anonymous'


class Proposal(object):

    @contract(validity='seq[2](int|None)', who='str')
    def __init__(self, validity, who):
        self.validity = tuple(validity)
        self.who = who

    def __eq__(self, b):
        return (self.validity == b.validity) and (self.who == b.who)

    def __hash__(self):
        return hash((self.validity, self.who))

    def __repr__(self):
        s = interval_friendly(self.validity)
        s += ' proposed by %r' % self.who
        return s

    @staticmethod
    def strictly_dominates(a, b, t=None):
        if t is None: t = time.time()
        ok1 = interval_strictly_dominates(a.validity, b.validity, t) and \
              who_dominates(a.who, b.who)
        ok2 = interval_dominates(a.validity, b.validity, t) and \
              who_strictly_dominates(a.who, b.who)
        return ok1 or ok2

    @staticmethod
    def dominates(a, b, t=None):
        if t is None: t = time.time()
        return interval_dominates(a.validity, b.validity, t) and \
              who_dominates(a.who, b.who)


def who_strictly_dominates(who1, who2):
    return who2 == ANONYMOUS and who1 != ANONYMOUS


def who_dominates(who1, who2):
    return (who1 == who2) or who_strictly_dominates(who1, who2)

#def signatures_dominate(s1, s2):
#    return s1.issuperset(s2)
#
#
#def signatures_strictly_dominate(s1, s2):
#    return s1.issuperset(s2) and (s1 != s2)


def interval_dominates(v1, v2, t=None):
    if t is None: t = time.time()
    return interval_strictly_dominates(v1, v2, t) or (v1 == v2)


def interval_strictly_dominates(v1, v2, t=None):
    if t is None: t = time.time()
    v1 = interval_cut(v1, t)
    v2 = interval_cut(v2, t)

    start_1, end_1 = v1
    start_2, end_2 = v2

    if end_1 is None: end_1 = float('inf')
    if end_2 is None: end_2 = float('inf')

    ok1 = (start_1 <= start_2) and (end_1 > end_2)
    ok2 = (start_1 < start_2) and (end_1 >= end_2)
    return ok1 or ok2


#    @contract(returns=Proposal)
def interval_cut(validity, t):
    ''' Returns a new interval cut
        at time t -- "let's forget about the past" '''
    t0, t1 = validity
    if t0 is None:
        t0_ = t
    else:
        t0_ = max(t0, t)
    return (t0_, t1)


@contract(proposal=Proposal, proposals='set($Proposal)', returns='set($Proposal)')
def strictly_dominated_by_set(proposal, proposals):
    ''' Retuns a set of proposals that dominate this one. '''
    doms = set()
    for p in proposals:
        if Proposal.strictly_dominates(p, proposal):
            doms.add(p)
    return doms
#
#    if proposal.signatures:
#        doms = set()
#        # all signatures must be dominated
#        for s in proposal.signatures:
#            p2 = Proposal(proposal.validity, [s])
#            doms_s = set()
#            for existing in proposals:
#                if Proposal.strictly_dominates(existing, p2):
#                    doms_s.add(existing)
#            if not doms_s:
#                # p2 was not dominated
#                return set()
#            doms.update(doms_s)
#        return doms
#    else:
#        doms = set()
#        for existing in proposals:
#            if Proposal.strictly_dominates(existing, proposal):
#                doms.add(existing)
#        return doms

#
#
#def strictly_dominates(validity1, validity2, t):
#    start_1, end_1 = validity1
#    start_2, end_2 = validity2
#
#    if start_1 is None:
#        start_1 = 0
#    if start_2 is None:
#        start_2 = 0
#    if end_1 is None:
#        end_1 = t * 2
#    if end_2 is None:
#        end_2 = t * 2
#    start_1 = max(t, start_1)
#    start_2 = max(t, start_2)
#    ok1 = (start_1 <= start_2) and (end_1 > end_2)
#    ok2 = (start_1 < start_2) and (end_1 >= end_2)
#
#    return ok1 or ok2



from collections import OrderedDict, namedtuple
from duckietown_swarm.spawn_ipfs import get_ipfs_executable
import json
import os
import sys
import time

from contracts import contract, new_contract
from contracts.utils import check_isinstance, raise_wrapped
from contracts.utils import indent
from system_cmd import CmdException, system_cmd_result

from .utils import memoize_simple, tmpfile


class Timeout(Exception):
    pass


class InvalidHash(Exception):
    pass


class CouldNotPin(Exception):
    pass


IPFS_ls_entry = namedtuple('IPFS_ls_entry', 'name hash size')


@new_contract
def multihash(mh):
    check_isinstance(mh, str)
    return mh.startswith('Qm')


class IPFSInterface(object):

    def __init__(self, ipfs_path):
        if ipfs_path is None:
            d = os.path.expanduser('~/shared-logs/.cache/.ipfs')
            if os.path.exists(d):
                ipfs_path = d
            else:
                if 'IPFS_PATH' in os.environ:
                    ipfs_path = os.environ['IPFS_PATH']
                else:
                    ipfs_path = os.path.expanduser('~/.ipfs')
        self.ipfs_path = ipfs_path
        self.providers_ttl = 120
        self.providers_cache = {}
        self.debug = False

    @memoize_simple
    def get_executable(self):
        return get_ipfs_executable(self.ipfs_path)

    @memoize_simple
    def version(self):
        cmd = ['ipfs', '--version']
        res = self._cmd(cmd)
        version = res.stdout
        version = version.replace('ipfs', '')
        version = version.replace('version', '')
        version = version.strip()
        print('Detected version %r' % version)
        return version

    def _get_env(self):
        env = os.environ.copy()
        if self.ipfs_path is not None:
            env['IPFS_PATH'] = self.ipfs_path
        return env

    def _cmd(self, cmd):
        if cmd[0] == 'ipfs':
            cmd[0] = self.get_executable()

        if self.debug:
            print('$ %s' % " ".join(cmd))
        else:
            sys.stderr.write('.')
        t0 = time.time()
        res = system_cmd_result('.', cmd, raise_on_error=True, env=self._get_env())

        delta = time.time() - t0
        if self.debug:
            print('took %5.2f s: $ %s' % (delta, " ".join(cmd)))
        return res

    def get_keys(self):
        ipfs = self.get_executable()
        cmd = [ipfs, 'key', 'list', '-l']
        res = self._cmd(cmd)
        #print res.stdout.__repr__()
        lines = res.stdout.strip().split('\n')
        res = OrderedDict()
        for l in lines:
            tokens = l.strip().split(' ')
            assert len(tokens) == 2, tokens
            res[tokens[1]] = tokens[0]
        return res

    def gen_key(self, name, key_type, key_size):
        cmd = ['ipfs', 'key', 'gen', '--type', key_type, '--size', str(key_size), name]
        res = self._cmd(cmd)
        ipfs_hash = res.stdout.strip()
        return ipfs_hash

    def publish(self, key_name, ipfs_hash, timeout='60s'):
        cmd = ['ipfs', 'name', 'publish', '--timeout', timeout, '--key', key_name, ipfs_hash]
        _res = self._cmd(cmd)

    def pin_add(self, mh, recursive=True, timeout=None):
        cmd = ['ipfs', 'pin', 'add']
        if recursive:
            cmd.append('-r')
        if timeout is not None:
            cmd.extend(['--timeout', timeout])
        cmd.append(mh)
        res = self._cmd(cmd)
        if res.ret != 0:
            raise CouldNotPin(str(res))

    def pin_ls(self):
        cmd = ['ipfs', 'pin', 'ls']
        res = self._cmd(cmd)
        recursive = set()
        indirect = set()
        for line in res.stdout.strip().split('\n'):
            tokens = line.split(' ')
            hashed = tokens[0]
            if tokens[1] == 'recursive':
                recursive.add(hashed)
            elif tokens[1] == 'indirect':
                indirect.add(hashed)
            else:
                assert False, line

        return recursive, indirect

    def dht_findprovs(self, ipfs_hash, timeout="1s"):
        if ipfs_hash in self.providers_cache:
            t, result = self.providers_cache[ipfs_hash]
            elapsed = time.time() - t
            if elapsed < self.providers_ttl:
                return result
        result = self._dht_findprovs(ipfs_hash, timeout)
        self.providers_cache[ipfs_hash] = time.time(), result
        return result

    def _dht_findprovs(self, ipfs_hash, timeout):
        cmd = ['ipfs', 'dht', 'findprovs', '--timeout', timeout, ipfs_hash]
        try:
            res = self._cmd(cmd)
        except CmdException as e:
            res = e.res
        options = res.stdout.strip().split('\n')
        options = [x for x in options if x]
        return options

    @memoize_simple
    def object_get(self, h, timeout=None):
        cmd = ['ipfs', 'object']
        if timeout is not None:
            cmd.extend([ '--timeout', timeout])
        cmd.extend(['get', h])

        try:
            res = self._cmd(cmd)
        except CmdException as e:
            raise InvalidHash(e.res.stderr)
        return res.stdout

    @memoize_simple
    def get(self, mh, timeout="1h"):
        with tmpfile('.ipfs_data') as f:
            cmd = ['ipfs', 'get', '--timeout', timeout, '-o', f, mh]
            try:
                _res = self._cmd(cmd)
            except CmdException as e:
                if e.res.ret == 1 and 'request canceled' in e.res.stderr:
                    msg = 'Could not get %s with timeout %s' % (mh, timeout)
                    raise_wrapped(Timeout, e, msg)
                print('exc %s %r' % (e.res.ret, e.res.stderr))
                raise
            data = open(f).read()
            return data

    def ls(self, h, timeout="10s"):
    #    {"Links":[{"Name":"FILE2","Hash":"QmUtkGLvPf63NwVzLPKPUYgwhn8ZYPWF6vKWN3fZ2amfJF","Size":14},
    # {"Name":"upload_info1.yaml","Hash":"QmeiuS7VWRaUQTmva3UMgggCCEDisgtLWPwxFnJ1kCGaDJ","Size":214}],"Data":"\u0008\u0001"}
        data = self.object_get(h, timeout=timeout)
        d = json.loads(data)
        entries = OrderedDict()
        for entry in d['Links']:
            entries[str(entry['Name'])] = \
                IPFS_ls_entry(str(entry['Name']), str(entry['Hash']), entry['Size'])
        return entries

    def add_bytes(self, s):
        return self.get_hash_for_bytes(s)

    def block_put(self, s):
        ipfs = self.get_executable()
        cmd = [ipfs, 'block', 'put']
        res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True, display_stdout=False, display_stderr=False,
                                write_stdin=s, env=self._get_env())
        token = res.stdout.strip()
        return token

    @memoize_simple
    def _get_ipfs_id_data(self):
        cmd = ['ipfs', 'id']
        res = self._cmd(cmd)

        data = json.loads(res.stdout.strip())
        return data

    def stats_bw(self):
        cmd = ['ipfs', 'stats', 'bw']
        res = self._cmd(cmd)
        return res.stdout.strip()

    @memoize_simple
    def ipfs_id(self):
        data = self._get_ipfs_id_data()
        return str(data['ID'])

    @memoize_simple
    @contract(returns='list(str)')
    def get_addresses(self):
        data = self._get_ipfs_id_data()
        if data['Addresses']:
            return map(str, data['Addresses'])
        else:
            return []

    def p2p_listener_open(self, name, address):
        cmd = ['ipfs', 'p2p', 'listener', 'ls']
        res = self._cmd(cmd)
        already = name in res.stdout
        if already:
            cmd = ['ipfs', 'p2p', 'listener', 'close', name]
            res = self._cmd(cmd)

        cmd = ['ipfs', 'p2p', 'listener', 'open', name, address]
        res = self._cmd(cmd)
        return json.loads(res.stdout.strip())

    def swarm_connect(self, peer, timeout='10s'):
        cmd = ['ipfs', 'swarm', '--timeout', timeout, 'connect', peer]
        res = self._cmd(cmd)
        print(res.stdout)

    def object_put(self, data):
        ipfs = self.get_executable()
        cmd = [ipfs, 'object', 'put']
        cwd = '.'
        res = system_cmd_result(cwd, cmd, raise_on_error=True,
                                write_stdin=data, env=self._get_env())
        hashed = res.stdout.split()[1]
        assert 'Qm' in hashed, hashed
        return hashed

    def add(self, s):
        ipfs = self.get_executable()
        cmd = [ipfs, 'add']
        cwd = '.'
        res = system_cmd_result(cwd, cmd, raise_on_error=True,
                                write_stdin=s, env=self._get_env())
        hashed = res.stdout.split()[1]
        assert 'Qm' in hashed, hashed
        return hashed

    get_hash_for_bytes = add

    def add_ipfs_dir(self, dirname):
        ipfs = self.get_executable()
        cmd = [ipfs, 'add', '-r', dirname]
        res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True,
                                env=self._get_env())
        s = res.stdout.strip()
        lines = s.split('\n')
        out = lines[-1].split(' ')
        if (len(out) < 3 or out[0] != 'added' or not out[1].startswith('Qm')):
            msg = 'Invalid output for ipds:\n%s' % indent(res.stdout, ' > ')
            raise Exception(msg)
        hashed = out[1]
        return hashed

    def get_tree_builder(self):
        return MakeIPFS3(self)


class MakeIPFS3(object):

    def __init__(self, ipfsi):
        self.ipfsi = ipfsi
        self.links = []
        self.total_file_size = 0

    def add_file(self, filename, mh, size):
        x = {'Name': filename, 'Hash': mh, "Size": size}
        self.links.append(x)
        self.total_file_size += size

    @contract(filename=str, s=str)
    def add_file_content(self, filename, s):
        # TODO: check filename
        mh = self.ipfsi.get_hash_for_bytes(s)
        self.add_file(filename, mh, len(s))

    def get_dag(self):
        result = {'Data': u"\u0008\u0001", 'Links': self.links}
        return result

    def total_size(self):
        return self.total_file_size

    def as_json(self):
        dag = self.get_dag()
        dag_json = json.dumps(dag)
        return dag_json

    @contract(returns='multihash')
    def get_hash(self):
        return self.ipfsi.object_put(self.as_json())

#
#
#def ipfs_pubsub_send(topic, message):
#    cmd = ['ipfs', 'pubsub', 'pub', topic, message + '\n']
#    system_cmd_result('.', cmd, raise_on_error=True)



from duckietown_swarm.dcache import BucketDataItem
from duckietown_swarm.proposals import Proposal
import time

from comptests.registrar import run_module_tests, comptest  #@UnresolvedImport


@comptest
def test_intervals1():
    p1 = Proposal([10, 20], [])
    p2 = p1
    assert Proposal.dominates(p1, p2)
    assert not Proposal.strictly_dominates(p1, p2)


@comptest
def test_intervals2():
    p1 = Proposal([10, 21], [])
    p2 = Proposal([10, 20], [])
    assert Proposal.strictly_dominates(p1, p2)


@comptest
def test_intervals3():
    p1 = Proposal([10, 20], ['sign1'])
    p2 = Proposal([10, 20], [])
    assert Proposal.strictly_dominates(p1, p2)


@comptest
def test_intervals4():
    p1 = Proposal([10, 20], ['sign1'])
    p2 = Proposal([10, 20], ['sign2'])
    assert not Proposal.dominates(p1, p2)


@comptest
def test_di1():
    from_channel = 'c'
    bdi = BucketDataItem()
    bdi.propose([10, 20], [], from_channel)
    bdi.propose([10, 20], [], from_channel)
    print bdi.summary()
    print bdi.proposals_propose
    assert len(bdi.proposals_propose) == 1


@comptest
def test_di2():
    from_channel = 'c'
    bdi = BucketDataItem()
    bdi.propose([10, 20], [], from_channel)
    bdi.propose([10, 15], [], from_channel)
    print bdi.proposals_propose
    assert len(bdi.proposals_propose) == 1


@comptest
def test_di3():
    t = int(time.time())
    from_channel = 'c'
    bdi = BucketDataItem()
    bdi.propose([t + 10, t + 20], [], from_channel)
    bdi.propose([t + 10, t + 20], ['frank'], from_channel)
    assert len(bdi.proposals_propose) == 1
    bdi.propose([t + 10, t + 30], ['frank'], from_channel)
    assert len(bdi.proposals_propose) == 1
    bdi.propose([t + 10, t + 30], ['john'], from_channel)
    assert len(bdi.proposals_propose) == 2


if __name__ == '__main__':
    run_module_tests()



import sys

from contracts import contract
from system_cmd import CmdException

from .brain import DistributedRepoSwarm
from .ipfs_utils import Timeout
from .irc2 import MessagingInterface

seen_mh = set()
seen = set()


@contract(mi=MessagingInterface)
def read_nodes_summaries(mi):
    timeout = '60s'
    ipfsi = mi.get_ipfsi()

    while True:
#        print('waiting for summary')
        envelopes = [mi.get_next_for_me()]
#
#        n = len(envelopes)
#        if n > 1:
#            print('ignoring %s messages and process the last' % (n - 1))
#            envelopes = [envelopes[-1]]

        for envelope in envelopes:
#
#            if envelope.maddr_from == mi.MADDR_BRAIN:
#                print('skipping this from ourselves: %s' % envelope)
#                continue

#            print envelope

            mh = envelope.contents
            if mh in seen_mh:
                continue

            seen_mh.add(mh)
            data_mh = mh + '/machines.txt'
            print('getting %s' % data_mh)

            try:
                data = ipfsi.get(data_mh, timeout=timeout)
            except Timeout:
                print('timeout for %s' % data_mh)
                continue
            except CmdException as e:
                print(str(e))
                continue

            data = data.strip()

            lines = data.split('\n')

            i = 0
            for msg in lines:
                if msg in seen:
                    continue
                else:
#                    print('NEW: %s' % msg)
                    seen.add(msg)
                    i += 1
                    from_channel = '/ipfs/' + mh
                    mi.send_to_brain(from_channel, msg)

            new_ones = i
            print('read %s catchup messages: %s new ones' % (len(lines), new_ones))


def go_offline():
    s = sys.stdin.read()
    dc = read_lines(s)

    show = ['networks', 'peer', 'trusted', 'admin', 'files', 'verified', 'safe', 'summaries',
            'pri']
    for k in show:
        for x in dc.query(k):
            print('%s: %s' % (k, x))

#initialized bucket networks
#initialized bucket peer
#initialized bucket trusted
#initialized bucket admin
#initialized bucket files
#initialized bucket verified
#initialized bucket safe
#initialized bucket summaries
#initialized bucket pri


def read_lines(s):
    lines = s.split('\n')
    dc = DistributedRepoSwarm()
    from_channel = 'stdin'
    i = 0
    for msg in lines:
        dc.process(msg, from_channel)
        i += 1
    print('processed: %s' % i)
    return dc


if __name__ == '__main__':
    go_offline()




from duckietown_swarm.irc2 import prefix_match

from comptests.registrar import run_module_tests, comptest


def should_match(address, queue, rest0):
    a, rest = prefix_match(address, queue)
    print address, queue
    print a, rest
    assert a
    assert rest == rest0, (a, rest)


def should_not_match(address, queue):
    a, rest = prefix_match(address, queue)
    print address, queue
    print a, rest
    assert not a


@comptest
def test_routing1():

    should_match('/irc/host/channel', '/irc/host/channel', '')
    should_match('/irc/host/channel/b', '/irc/host/channel', '/b')
    should_match('/irc/*/channel', '/irc/host/channel', '')


@comptest
def test_routing2():
    should_match('/irc/*/channel/b', '/irc/host/channel', '/b')


@comptest
def test_routing3():
    a = '/dswarm/QmQRxFjxFdj5pgUrriNZ1jFeTKwhrc5hTFAE9R3MHbnfrr/brain'
    pipe = '/dswarm/QmQRxFjxFdj5pgUrriNZ1jFeTKwhrc5hTFAE9R3MHbnfrr/pinner'
    # with rest '/brain'
    should_not_match(a, pipe)


if __name__ == '__main__':
    run_module_tests()



import time

from system_cmd import system_cmd_result

from .dcache import create_propose_message


def pubsub_friendship(mi):
    ipfsi = mi.get_ipfsi()

    ## Read from the queue
    ID = ipfsi.ipfs_id()

    # signal that we are here
    token = ipfsi.block_put('duckiebots\n')

    found = set()
    while True:
        # find other duckiebots
        hosts = ipfsi.dht_findprovs(token, timeout='30s')

        # print('Found %s Duckiebots in the hidden botnet.' % len(hosts))
        for h in hosts:
            # do not connect to self
            if h == ID: continue
            if h in found:
                continue

            # publicize peer
            t = time.time()
            interval = (t, t + 9 * 60)
            mi.send_to_brain('', create_propose_message('peer', h, interval))

            found.add(h)

        time.sleep(30)


def connecting_to_peer(mi):
    ipfsi = mi.get_ipfsi()
    ID = ipfsi.ipfs_id()

    while True:
        envelope = mi.get_next_for_me()
        h = envelope.contents
        ipfs = ipfsi.get_executable()
        cmd = [ipfs, 'swarm', 'connect', '--timeout', '30s', h]
        res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=False,
                              display_stdout=False, display_stderr=False,
                              env=ipfsi._get_env())
        ok = res.ret == 0
        from_where = envelope.maddr_from
        print('Connecting to new friend %s (%s): ' % (h, from_where) + ("OK" if ok else "(not possible): " + str(res)))
        if ok:
            interval = (time.time(), time.time() + 9 * 60)
            connection = '%s:%s' % (ID, h)
            msg = create_propose_message('connections', connection, interval)
            mi.send_to_brain('', msg)





from multiprocessing import Queue
import socket
from threading import Thread

from contracts.utils import check_isinstance

from .dcache import Envelope
from .tcpendpoint import MakeLines


class AsyncLineBasedConnection():

    def __init__(self, host, port):
        addr = (host, port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(addr)
        print('Connected to %s:%s' % (addr))

        self.q_in = Queue()
        self.q_out = Queue()

        address = '/ip4/%s/tcp/%s/lswarm' % (host, port)
        thread1 = Thread(target=read_from_socket, args=(s, self.q_in, address))
        thread1.daemon = True
        thread1.start()
        thread2 = Thread(target=write_to_socket, args=(s, self.q_out,))
        thread2.daemon = True
        thread2.start()

    def get_q_inbound(self):
        return self.q_in

    def get_q_outbound(self):
        return self.q_out


def read_from_socket(socket, q_in, address):
    ml = MakeLines()
    while True:
        chars = socket.recv(1)
        ml.push(chars)
        for line in ml.get_lines():
            env = Envelope.from_json(line)
            env.maddr_from = address + env.maddr_from
            q_in.put(env)


def write_to_socket(socket, q_out):
    while True:
        env = q_out.get()
        check_isinstance(env, Envelope)
        line = env.to_json()
        print('writing %s' % env.contents)
        socket.sendall(line + '\n')





from duckietown_swarm.dcache_wire import create_propose_message, \
    interpret_message, create_dismiss_message, create_signature_message

from comptests.registrar import run_module_tests, comptest


@comptest
def test_yaml_propose():
    buckets = ('one', 'two')
    data = 'mydata'
    validity_interval = (0, None)
    msg = create_propose_message(buckets=buckets, data=data, validity=validity_interval)
    s = interpret_message(msg)
    assert s.buckets == buckets, s.buckets
    assert s.validity == validity_interval, s
    assert s.data == data


@comptest
def test_yaml_propose2():
    buckets = 'thebucket'
    data = 'mydata'
    validity_interval = (0, None)
    msg = create_propose_message(buckets=buckets, data=data, validity=validity_interval)
    s = interpret_message(msg)
    assert s.buckets == (buckets,), s.buckets
    assert s.validity == validity_interval, s
    assert s.data == data


@comptest
def test_yaml_dismiss():
    buckets = ('one', 'two')
    data = 'mydata'
    validity_interval = (0, None)
    msg = create_dismiss_message(buckets=buckets, data=data, validity=validity_interval)
    s = interpret_message(msg)
    assert s.buckets == buckets, s.buckets
    assert s.validity == validity_interval, s
    assert s.data == data


@comptest
def test_yaml_signature():
    signer = 'me'
    data_hash = 'mydata'
    signature = 'signature'
    msg = create_signature_message(data_hash=data_hash, signer=signer, signature=signature)
    s = interpret_message(msg)
    assert s.signer == signer, s
    assert s.signature == signature, s
    assert s.data == data_hash


if __name__ == '__main__':
    run_module_tests()



#! /usr/bin/env python

import time

from .utils import duration_compact


def time_friendly(delta):
    if delta > 0:
        return duration_compact(delta)
    else:
        return duration_compact(-delta) + ' ago'


def interval_friendly(interval):
    t0, t1 = interval
    if t0:
        d0 = time_friendly(time.time() - t0)
    else:
        d0 = "forever"
    if t1:
        d1 = time_friendly(t1 - time.time())
    else:
        d1 = 'forever'
    return '[%s to %s]' % (d0, d1)


def in_interval(interval, at):
    t0, t1 = interval
    ok1 = t0 is None or t0 <= at
    ok2 = t1 is None or at <= t1
    return ok1 and ok2



from collections import OrderedDict, namedtuple
import copy
import json
import time

from contracts import contract
from contracts.utils import check_isinstance, raise_wrapped

MSG_PROPOSE = 'propose'
MSG_DISMISS = 'dismiss'
MSG_SIGNATURE = 'signature'
MSG_PING = 'ping'
MSG_PONG = 'pong'
MSG_REQ = 'req'

FIELD_BUCKET = 'bucket'
FIELD_MTYPE = 'mtype'
FIELD_VALIDITY = 'validity'
FIELD_PROPOSE_DATA = 'data'
FIELD_SIGN_DATA = 'data'
FIELD_SIGNER = 'signer'
FIELD_SIGNATURE = 'signature'
FIELD_PATTERNS = 'patterns'

MsgPropose = namedtuple('MsgPropose', 'buckets data validity')
MsgDismiss = namedtuple('MsgDismiss', 'buckets data validity')
MsgSignature = namedtuple('MsgSignature', 'signer data signature')
MsgPing = namedtuple('MsgPing', 'validity')
MsgPong = namedtuple('MsgPong', 'validity')
MsgRequest = namedtuple('MsgRequest', 'patterns validity')
# general purpose commands
MSG_COMMAND = 'command'
FIELD_COMMAND = 'command'
FIELD_PARAMETERS = 'parameters'
MsgCommand = namedtuple('MsgCommand', 'command parameters validity')


@contract(buckets='str|seq(str)', data=str, validity='None|seq[2]', returns=str)
def create_propose_message(buckets, data, validity=None):
    '''
        Done:

            {'mtype': 'propose', 'data': data, 'validity': [0, null]}
    '''
    if isinstance(buckets, str):
        buckets = (buckets,)

    if validity is None:
        validity = None, None

    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_PROPOSE
    res[FIELD_BUCKET] = list(buckets)
    res[FIELD_PROPOSE_DATA] = data
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


def create_pong_message(delta=5):
    t = time.time()
    validity = [t, t + delta]
    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_PONG
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


def create_ping_message(delta=5):
    t = time.time()
    validity = [t, t + delta]
    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_PING
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


def create_command_message(command, parameters, validity):
    res = OrderedDict()
    res[FIELD_COMMAND] = command
    res[FIELD_PARAMETERS] = parameters
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


@contract(d='dict')
def interpret_command_message(d):
    d2 = copy.deepcopy(d)
    assert d2.pop(FIELD_MTYPE) == MSG_COMMAND

    command = str(d2.pop(FIELD_COMMAND))
    parameters = (d2.pop(FIELD_PARAMETERS))
    validity = tuple(d2.pop(FIELD_VALIDITY))
    if d2:
        msg = 'Additional fields %s in %s' % (list(d2), d)
        raise CouldNotInterpret(msg)
    return MsgCommand(command=command, parameters=parameters, validity=validity)


@contract(patterns='seq(str)')
def create_request_message(patterns, validity):
    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_REQ
    res[FIELD_PATTERNS] = patterns
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


@contract(d='dict')
def interpret_request_message(d):
    d2 = copy.deepcopy(d)
    assert d2.pop(FIELD_MTYPE) == MSG_REQ

    patterns = tuple(d2.pop(FIELD_PATTERNS))
    validity = tuple(d2.pop(FIELD_VALIDITY))
    if d2:
        msg = 'Additional fields %s in %s' % (list(d2), d)
        raise CouldNotInterpret(msg)
    return MsgRequest(patterns=patterns, validity=validity)


@contract(buckets='str|seq(str)', data=str, validity='seq[2]', returns=str)
def create_dismiss_message(buckets, data, validity):
    '''
        Done:

            {'mtype': 'dismiss', 'data': data, 'validity': [0, null]}
    '''
    if isinstance(buckets, str):
        buckets = (buckets,)

    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_DISMISS
    res[FIELD_BUCKET] = list(buckets)
    res[FIELD_PROPOSE_DATA] = data
    res[FIELD_VALIDITY] = _format_validity(validity)
    s = json.dumps(res).encode()
    return s


def _format_validity(validity):

    def intornone(x):
        return None if x is None else int(x)

    return [intornone(validity[0]), intornone(validity[1])]


@contract(d='dict')
def interpret_propose_message(d):
    d2 = copy.deepcopy(d)
    assert d2.pop(FIELD_MTYPE) == MSG_PROPOSE

    buckets = tuple(map(str, d2.pop(FIELD_BUCKET)))
    data = str(d2.pop(FIELD_PROPOSE_DATA))
    validity = tuple(d2.pop(FIELD_VALIDITY))
    if d2:
        msg = 'Additional fields %s in %s' % (list(d2), d)
        raise CouldNotInterpret(msg)
    return MsgPropose(buckets=buckets, data=data, validity=validity)


@contract(d='dict')
def interpret_dismiss_message(d):
    d2 = copy.deepcopy(d)
    assert d2.pop(FIELD_MTYPE) == MSG_DISMISS

    buckets = tuple(map(str, d2.pop(FIELD_BUCKET)))
    data = str(d2.pop(FIELD_PROPOSE_DATA))
    validity = tuple(d2.pop(FIELD_VALIDITY))
    if d2:
        msg = 'Additional fields %s in %s' % (list(d2), d)
        raise CouldNotInterpret(msg)
    return MsgDismiss(buckets=buckets, data=data, validity=validity)


@contract(d='dict')
def interpret_ping_message(d):
    validity = tuple(d[FIELD_VALIDITY])
    return MsgPing(validity=validity)


@contract(d='dict')
def interpret_pong_message(d):
    validity = tuple(d[FIELD_VALIDITY])
    return MsgPong(validity=validity)


class CouldNotInterpret(Exception):
    pass


@contract(s=str)
def interpret_message(s):
    check_isinstance(s, str)
    cases = {
        MSG_PROPOSE: interpret_propose_message,
        MSG_DISMISS: interpret_dismiss_message,
        MSG_SIGNATURE: interpret_signature_message,
        MSG_PING: interpret_ping_message,
        MSG_PONG: interpret_pong_message,
        MSG_REQ: interpret_request_message,
        MSG_COMMAND: interpret_command_message,
    }
    try:
        d = json.loads(s)
    except ValueError as e:
        msg = 'Could not read JSON:\n\n   %s' % s
        raise_wrapped(CouldNotInterpret, e, msg)
    check_isinstance(d, dict)
    try:
        mtype = d['mtype']
    except KeyError:
        msg = 'Cannot interpret message: %r' % s
        raise CouldNotInterpret(msg)
    if mtype in cases:
        return cases[mtype](d)
    else:
        msg = 'Invalid type %r' % mtype
        raise CouldNotInterpret(msg)


@contract(data_hash='str', signer='str', signature=str, returns=str)
def create_signature_message(data_hash, signer, signature):
    '''
        data_mhash

        Result:

            {'mtype': 'sign', 'hash': data, 'signer': hash, 'signature': signature}

    '''
    res = OrderedDict()
    res[FIELD_MTYPE] = MSG_SIGNATURE
    res[FIELD_SIGN_DATA] = data_hash
    res[FIELD_SIGNER] = signer
    res[FIELD_SIGNATURE] = signature
    s = json.dumps(res).encode()
    return s


@contract(d='dict')
def interpret_signature_message(d):
    d2 = copy.deepcopy(d)
    assert d2.pop(FIELD_MTYPE) == MSG_SIGNATURE

    signature = str(d2.pop(FIELD_SIGNATURE))
    data = str(d2.pop(FIELD_SIGN_DATA))
    signer = str(d2.pop(FIELD_SIGNER))
    if d2:
        msg = 'Additional fields %s in %s' % (list(d2), d)
        raise CouldNotInterpret(msg)
    return MsgSignature(signature=signature, data=data, signer=signer)


@contract(message='tuple(str, str)')
def put_in_queue(queue, message):
    check_isinstance(message, tuple)
    print('put_in_queue %s %s' % message)
    queue.put(message)



from contextlib import contextmanager
import math
import os
from tempfile import NamedTemporaryFile
import time

import base58
from contracts.utils import indent
from decorator import decorator
from system_cmd.meat import system_cmd_result


def get_sha256_base58(contents):
    import hashlib
    m = hashlib.sha256()
    m.update(contents)
    s = m.digest()
    return base58.b58encode(s)


def get_sha256_for_file_hex(filename):
    import hashlib
    m = hashlib.sha256()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)
    return m.hexdigest()


def yaml_dump_omaps(s):
    from ruamel import yaml
    res = yaml.dump(s, Dumper=yaml.RoundTripDumper, allow_unicode=False)
    return res


def yaml_load_omaps(s):
    from ruamel import yaml
    res = yaml.load(s, Loader=yaml.UnsafeLoader)
    return res


@contextmanager
def tmpfile(suffix):
    ''' Yields the name of a temporary file '''
    temp_file = NamedTemporaryFile(suffix=suffix)
    try:
        yield temp_file.name
    finally:
        temp_file.close()

# -*- coding: utf-8 -*-


__all__ = [
    'memoize_simple',
]


def memoize_simple(obj):
    cache = obj.cache = {}

    def memoizer(f, *args):
        key = (args)
        if key not in cache:
            cache[key] = f(*args)
        assert key in cache

        try:
            cached = cache[key]
            return cached
        except ImportError:  # pragma: no cover  # impossible to test
            del cache[key]
            cache[key] = f(*args)
            return cache[key]

            # print('memoize: %s %d storage' % (obj, len(cache)))

    return decorator(memoizer, obj)


def pretty_print_dictionary(d):
    lengths = [len(k) for k in d.keys()]
    if not lengths:
        return 'Empty.'

    s = ""
    for k, v in d.items():
        if isinstance(k, tuple):
            k = k.__repr__()
        if s:
            s += '\n\n'
        s += k
        s += '\n\n' + indent(str(v), '  ')
    return s


def get_all_available(queue):
    ''' Gets all that are available currently. '''
    from Queue import Empty
    res = []
    while True:
        try:
            x = queue.get(block=False)
            res.append(x)
        except Empty:
            break
    return res


def get_at_least_one(queue, timeout):
    """ This waits for one, and then waits at most timeout. """
    from Queue import Empty
    s = []
    t0 = time.time()
    while True:
        try:
            block = len(s) == 0
            x = queue.get(block=block, timeout=timeout)
            s.append(x)

            if time.time() < t0 + timeout:
                wait = t0 + timeout - time.time()
                time.sleep(wait)
        except Empty:
            break

    T = time.time() - t0
    if False:
        print('Got %s objects after %.1f s (timeout %.1f s)' % (len(s), T, timeout))
    return s


def friendly_time_since(t0):
    if t0 == 0:
        return 'never'
    delta = time.time() - t0
    return duration_compact(delta)


def duration_compact(seconds):
    if seconds < 0:
        raise ValueError(seconds)
    seconds = int(math.ceil(seconds))
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    years, days = divmod(days, 365.242199)

    minutes = int(minutes)
    hours = int(hours)
    days = int(days)
    years = int(years)

    duration = []
    if years > 0:
        duration.append('%dy' % years)
    else:
        if days > 0:
            duration.append('%dd' % days)
        if (days < 3) and (years == 0):
            if hours > 0:
                duration.append('%dh' % hours)
            if (hours < 3) and (days == 0):
                if minutes > 0:
                    duration.append('%dm' % minutes)
                if (minutes < 3) and (hours == 0):
                    if seconds > 0:
                        duration.append('%ds' % seconds)

    return ' '.join(duration)


class DoAtInterval():

    def __init__(self, every_s):
        self.last = 0
        self.every = every_s

    def its_time(self):
        delta = time.time() - self.last
        if delta > self.every:
            self.last = time.time()
            return True
        else:
            return False


class MakeLines():

    def __init__(self):
        self.data = ''
        self.lines_in = []

    def get_lines(self):
        for l in self.lines_in:
            yield l
        self.lines_in = []

    def push(self, ks):
        for k in ks:
            if k == '\n':
                self.lines_in.append(self.data)
                self.data = ''
            else:
                self.data += k


class LineSplitter(object):
    """ A simple utility to split an incoming sequence of chars
        in lines. Push characters using append_chars() and
        get the completed lines using lines(). """

    def __init__(self):
        self.current = ''
        self.current_lines = []

    def append_chars(self, s):
        # TODO: make this faster
        s = str(s)
        for char in s:
            if char == '\n':
                self.current_lines.append(self.current)
                self.current = ''
            else:
                self.current += char

    def lines(self):
        """ Returns a list of line; empties the buffer """
        l = list(self.current_lines)
        self.current_lines = []
        return l


def now_until(delta):
    t = time.time()
    return [t, t + delta]


def download_url_to_file(url, filename):
    print('Download from %s' % (url))
    tmp = filename + '.tmp_download_file'
    cmd = [
        'wget',
        '-O',
        tmp,
        url
    ]
    d = os.path.dirname(tmp)
    if not os.path.exists(d):
        os.makedirs(d)
    res = system_cmd_result(cwd='.',
                          cmd=cmd,
                          display_stdout=True,
                          display_stderr=True,
                          raise_on_error=True,
                          write_stdin='',
                          capture_keyboard_interrupt=False,
                          env=None)

    if not os.path.exists(tmp) and not os.path.exists(filename):
        msg = 'Downloaded file does not exist but wget did not give any error.'
        msg += '\n url: %s' % url
        msg += '\n downloaded to: %s' % tmp
        msg += '\n' + indent(str(res), ' | ')
        d = os.path.dirname(tmp)
        r = system_cmd_result(d, ['ls', '-l'], display_stdout=False,
                          display_stderr=False,
                          raise_on_error=True)
        msg += '\n Contents of the directory:'
        msg += '\n' + indent(str(r.stdout), ' | ')
        raise Exception(msg)

    if not os.path.exists(filename):
        os.rename(tmp, filename)

#    logger.info('-> %s' % friendly_path(filename))
